@@ -1,7 +1,7 @@
 //
 // ZPL driver for LPrint, a Label Printer Application
 //
-// Copyright © 2019-2021 by Michael R Sweet.
+// Copyright © 2019-2024 by Michael R Sweet.
 // Copyright © 2007-2019 by Apple Inc.
 // Copyright © 2001-2007 by Easy Software Products.
 //
@@ -9,10 +9,6 @@
 // information.
 //
 
-//
-// Include necessary headers...
-//
-
 #include "lprint.h"
 
 
@@ -53,6 +49,7 @@
 
 typedef struct lprint_zpl_s		// ZPL driver data
 {
+  lprint_dither_t dither;		// Dither buffer
   unsigned char	*comp_buffer;		// Compression buffer
   unsigned char *last_buffer;		// Last line
   int		last_buffer_set;	// Is the last line set?
@@ -185,6 +182,7 @@ static bool	lprint_zpl_rstartjob(pappl_job_t *job, pappl_pr_options_t *options,
 static bool	lprint_zpl_rstartpage(pappl_job_t *job, pappl_pr_options_t *options, pappl_device_t *device, unsigned page);
 static bool	lprint_zpl_rwriteline(pappl_job_t *job, pappl_pr_options_t *options, pappl_device_t *device, unsigned y, const unsigned char *line);
 static bool	lprint_zpl_status(pappl_printer_t *printer);
+static bool	lprint_zpl_update_reasons(pappl_printer_t *printer, pappl_job_t *job, pappl_device_t *device);
 
 
 //
@@ -208,7 +206,7 @@ lprintZPL(
   data->rstartpage_cb = lprint_zpl_rstartpage;
   data->rwriteline_cb = lprint_zpl_rwriteline;
   data->status_cb     = lprint_zpl_status;
-  data->format        = "application/vnd.zebra-zpl";
+  data->format        = LPRINT_ZPL_MIMETYPE;
 
   data->num_resolution = 1;
 
@@ -239,9 +237,8 @@ lprintZPL(
     data->num_media = (int)(sizeof(lprint_zpl_2inch_media) / sizeof(lprint_zpl_2inch_media[0]));
     memcpy(data->media, lprint_zpl_2inch_media, sizeof(lprint_zpl_2inch_media));
 
-    papplCopyString(data->media_default.size_name, "oe_2x3-label_2x3in", sizeof(data->media_default.size_name));
-    data->media_default.size_width  = 2 * 2540;
-    data->media_default.size_length = 3 * 2540;
+    papplCopyString(data->media_ready[0].size_name, "oe_2x3-label_2x3in", sizeof(data->media_ready[0].size_name));
+    papplCopyString(data->media_ready[0].type, "labels", sizeof(data->media_ready[0].type));
   }
   else
   {
@@ -249,11 +246,12 @@ lprintZPL(
     data->num_media = (int)(sizeof(lprint_zpl_4inch_media) / sizeof(lprint_zpl_4inch_media[0]));
     memcpy(data->media, lprint_zpl_4inch_media, sizeof(lprint_zpl_4inch_media));
 
-    papplCopyString(data->media_default.size_name, "na_index-4x6_4x6in", sizeof(data->media_default.size_name));
-    data->media_default.size_width  = 4 * 2540;
-    data->media_default.size_length = 6 * 2540;
+    papplCopyString(data->media_ready[0].size_name, "na_index-4x6_4x6in", sizeof(data->media_ready[0].size_name));
+    papplCopyString(data->media_ready[0].type, "labels", sizeof(data->media_ready[0].type));
   }
 
+  data->media_ready[0].tracking = PAPPL_MEDIA_TRACKING_GAP;
+
   data->bottom_top = data->left_right = 1;
 
   data->num_source = 1;
@@ -261,26 +259,15 @@ lprintZPL(
 
   data->top_offset_supported[0] = -1500;
   data->top_offset_supported[1] = 1500;
-  data->tracking_supported      = PAPPL_MEDIA_TRACKING_MARK | PAPPL_MEDIA_TRACKING_WEB | PAPPL_MEDIA_TRACKING_CONTINUOUS;
+  data->tracking_supported      = PAPPL_MEDIA_TRACKING_MARK | PAPPL_MEDIA_TRACKING_WEB | PAPPL_MEDIA_TRACKING_GAP | PAPPL_MEDIA_TRACKING_CONTINUOUS;
 
   data->num_type = 3;
   data->type[0]  = "continuous";
   data->type[1]  = "labels";
   data->type[2]  = "labels-continuous";
 
-  data->media_default.bottom_margin = data->bottom_top;
-  data->media_default.left_margin   = data->left_right;
-  data->media_default.right_margin  = data->left_right;
-  papplCopyString(data->media_default.source, "main-roll", sizeof(data->media_default.source));
-  data->media_default.top_margin = data->bottom_top;
-  data->media_default.top_offset = 0;
-  data->media_default.tracking   = PAPPL_MEDIA_TRACKING_MARK;
-  papplCopyString(data->media_default.type, "labels", sizeof(data->media_default.type));
-
-  data->media_ready[0] = data->media_default;
-
   data->mode_configured = PAPPL_LABEL_MODE_TEAR_OFF;
-  data->mode_configured = PAPPL_LABEL_MODE_APPLICATOR | PAPPL_LABEL_MODE_CUTTER | PAPPL_LABEL_MODE_CUTTER_DELAYED | PAPPL_LABEL_MODE_KIOSK | PAPPL_LABEL_MODE_PEEL_OFF | PAPPL_LABEL_MODE_PEEL_OFF_PREPEEL | PAPPL_LABEL_MODE_REWIND | PAPPL_LABEL_MODE_RFID | PAPPL_LABEL_MODE_TEAR_OFF;
+  data->mode_supported = PAPPL_LABEL_MODE_APPLICATOR | PAPPL_LABEL_MODE_CUTTER | PAPPL_LABEL_MODE_CUTTER_DELAYED | PAPPL_LABEL_MODE_KIOSK | PAPPL_LABEL_MODE_PEEL_OFF | PAPPL_LABEL_MODE_PEEL_OFF_PREPEEL | PAPPL_LABEL_MODE_REWIND | PAPPL_LABEL_MODE_RFID | PAPPL_LABEL_MODE_TEAR_OFF;
 
   data->tear_offset_configured   = 0;
   data->tear_offset_supported[0] = -1500;
@@ -446,6 +433,10 @@ lprint_zpl_printfile(
     return (false);
   }
 
+  // Update status...
+  lprint_zpl_update_reasons(papplJobGetPrinter(job), job, device);
+
+  // Copy print data...
   while ((bytes = read(fd, buffer, sizeof(buffer))) > 0)
   {
     if (papplDeviceWrite(device, buffer, (size_t)bytes) < 0)
@@ -459,6 +450,9 @@ lprint_zpl_printfile(
 
   papplJobSetImpressionsCompleted(job, 1);
 
+  // Update status...
+  lprint_zpl_update_reasons(papplJobGetPrinter(job), job, device);
+
   return (true);
 }
 
@@ -503,6 +497,8 @@ lprint_zpl_rendpage(
 
   (void)page;
 
+  lprint_zpl_rwriteline(job, options, device, options->header.cupsHeight, NULL);
+
   papplDevicePrintf(device, "^XA\n^POI\n^PW%u\n^LH0,0\n^LT%d\n", options->header.cupsWidth, options->media.top_offset * options->printer_resolution[1] / 2540);
 
   if (options->media.type[0] && strcmp(options->media.type, "labels"))
@@ -526,13 +522,19 @@ lprint_zpl_rendpage(
   else
     papplDevicePuts(device, "^MTD\n");	// Direct thermal
 
-  papplDevicePrintf(device, "^PQ%d, 0, 0, N\n", options->copies);
+  papplDevicePrintf(device, "^PQ1, 0, 0, N\n");
   papplDevicePuts(device, "^FO0,0^XGR:LPRINT.GRF,1,1^FS\n^XZ\n");
   papplDevicePuts(device, "^XA\n^IDR:LPRINT.GRF^FS\n^XZ\n");
 
   if (options->finishings & PAPPL_FINISHINGS_TRIM)
     papplDevicePuts(device, "^CN1\n");
 
+  // Update status...
+  lprint_zpl_update_reasons(papplJobGetPrinter(job), job, device);
+
+  // Free memory and return...
+  lprintDitherFree(&zpl->dither);
+
   free(zpl->comp_buffer);
   free(zpl->last_buffer);
 
@@ -551,10 +553,12 @@ lprint_zpl_rstartjob(
     pappl_device_t     *device)		// I - Output device
 {
   pappl_pr_driver_data_t data;		// Driver data
+  int			darkness;	// Composite darkness value
   lprint_zpl_t	*zpl = (lprint_zpl_t *)calloc(1, sizeof(lprint_zpl_t));
 					// ZPL driver data
 
 
+  // Initialize driver data...
   papplJobSetData(job, zpl);
 
   papplPrinterGetDriverData(papplJobGetPrinter(job), &data);
@@ -598,8 +602,13 @@ lprint_zpl_rstartjob(
   else if (data.tear_offset_configured > 0)
     papplDevicePrintf(device, "~TA%03d\n", data.tear_offset_configured);
 
-  // printer-darkness
-  papplDevicePrintf(device, "~SD%02u\n", 30 * data.darkness_configured / 100);
+  // print-darkness / printer-darkness-configured
+  if ((darkness = options->print_darkness + options->darkness_configured) < 0)
+    darkness = 0;
+  else if (darkness > 100)
+    darkness = 100;
+
+  papplDevicePrintf(device, "~SD%02u\n", 30 * darkness / 100);
 
   return (true);
 }
@@ -619,25 +628,41 @@ lprint_zpl_rstartpage(
   lprint_zpl_t	*zpl = (lprint_zpl_t *)papplJobGetData(job);
 					// ZPL driver data
   int		ips;			// Inches per second
+  double	out_gamma = 1.0;	// Output gamma correction
 
 
   (void)page;
 
-  // print-darkness
-  papplDevicePrintf(device, "~MD%d\n", 30 * options->print_darkness / 100);
+  // Update status...
+  lprint_zpl_update_reasons(papplJobGetPrinter(job), job, device);
+
+  // Setup dither buffer...
+  if (options->header.HWResolution[0] == 300)
+    out_gamma = 1.2;
+  else if (options->header.HWResolution[0] == 600)
+    out_gamma = 1.44;
+
+  if (!lprintDitherAlloc(&zpl->dither, job, options, CUPS_CSPACE_K, out_gamma))
+    return (false);
 
   // print-speed
   if ((ips = options->print_speed / 2540) > 0)
     papplDevicePrintf(device, "^PR%d,%d,%d\n", ips, ips, ips);
 
   // Download bitmap...
-  papplDevicePrintf(device, "~DGR:LPRINT.GRF,%u,%u,\n", options->header.cupsHeight * options->header.cupsBytesPerLine, options->header.cupsBytesPerLine);
+  papplDevicePrintf(device, "~DGR:LPRINT.GRF,%u,%u,\n", zpl->dither.in_height * zpl->dither.out_width, zpl->dither.out_width);
 
   // Allocate memory for writing the bitmap...
-  zpl->comp_buffer     = malloc(2 * options->header.cupsBytesPerLine + 1);
-  zpl->last_buffer     = malloc(options->header.cupsBytesPerLine);
+  zpl->comp_buffer     = malloc(2 * zpl->dither.out_width + 1);
+  zpl->last_buffer     = malloc(zpl->dither.out_width);
   zpl->last_buffer_set = 0;
 
+  if (!zpl->comp_buffer || !zpl->last_buffer)
+  {
+    papplLogJob(job, PAPPL_LOGLEVEL_ERROR, "Unable to allocate compression buffers.");
+    return (false);
+  }
+
   return (true);
 }
 
@@ -666,18 +691,20 @@ lprint_zpl_rwriteline(
   static const unsigned char *hex = (const unsigned char *)"0123456789ABCDEF";
 					// Hex digits
 
-  (void)y;
+
+  if (!lprintDitherLine(&zpl->dither, y, line))
+    return (true);
 
   // Determine whether this row is the same as the previous line.
   // If so, output a ':' and return...
-  if (zpl->last_buffer_set && !memcmp(line, zpl->last_buffer, options->header.cupsBytesPerLine))
+  if (zpl->last_buffer_set && !memcmp(zpl->dither.output, zpl->last_buffer, zpl->dither.out_width))
   {
     papplDeviceWrite(device, ":", 1);
     return (true);
   }
 
   // Convert the line to hex digits...
-  for (ptr = line, compptr = zpl->comp_buffer, i = options->header.cupsBytesPerLine; i > 0; i --, ptr ++)
+  for (ptr = zpl->dither.output, compptr = zpl->comp_buffer, i = zpl->dither.out_width; i > 0; i --, ptr ++)
   {
     *compptr++ = hex[*ptr >> 4];
     *compptr++ = hex[*ptr & 15];
@@ -723,7 +750,7 @@ lprint_zpl_rwriteline(
 #endif // ZPL_COMPRESSION
 
   // Save this line for the next round...
-  memcpy(zpl->last_buffer, line, options->header.cupsBytesPerLine);
+  memcpy(zpl->last_buffer, zpl->dither.output, zpl->dither.out_width);
   zpl->last_buffer_set = 1;
 
   return (true);
@@ -739,12 +766,8 @@ lprint_zpl_status(
     pappl_printer_t *printer)		// I - Printer
 {
   pappl_device_t	*device;	// Connection to printer
-  char			line[1025],	// Line from printer
-			*lineptr;	// Pointer into line
+  char			line[1025];	// Line from printer
   ssize_t		bytes;		// Bytes read
-  unsigned		errors = 0,	// Detected errors
-			warnings = 0;	// Detected warnings
-  pappl_preason_t	reasons;	// "printer-state-reasons" values
   int			length = 0;	// Label length
   bool			ret = false;	// Return value
 
@@ -756,26 +779,101 @@ lprint_zpl_status(
   }
 
   // Get the printer status...
-  if (papplDevicePuts(device, "~HQES\n") < 0)
+  if (!lprint_zpl_update_reasons(printer, NULL, device))
+    goto done;
+
+  // Query host status...
+  if (papplDevicePuts(device, "~HS\n") < 0)
   {
-    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Unable to send HQES status command.");
+    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Unable to send HS status command.");
     goto done;
   }
 
   if ((bytes = papplDeviceRead(device, line, sizeof(line) - 1)) < 0)
   {
-    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Unable to read HQES status response.");
+    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Unable to read HS status response.");
     goto done;
   }
 
   line[bytes] = '\0';
 
+  papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "HS returned '%s'.", line);
+
+  if (sscanf(line + 1, "%*d,%*d,%*d,%d", &length) == 1 && length > 11)
+  {
+    // Auto-detect label length for ready media...
+    pappl_pr_driver_data_t	data;	// Driver data
+
+    papplPrinterGetDriverData(printer, &data);
+
+    // Round the length to the nearest dot and snap to the nearest 1/4"...
+    length = (2540 * length + data.y_resolution[0] / 2) / data.y_resolution[0];
+
+    if ((length % 635) <= 100)
+      length += length % 635;
+    else if ((length % 635) >= 535)
+      length += 635 - (length % 635);
+
+    // Lookup size
+    lprintMediaMatch(printer, 0, 0, length);
+  }
+
+  ret = true;
+
+  done:
+
+  papplPrinterCloseDevice(printer);
+
+  return (ret);
+}
+
+
+//
+// 'lprint_zpl_update_reasons()' - Update "printer-state-reasons" values.
+//
+
+static bool				// O - `true` on success, `false` on failure
+lprint_zpl_update_reasons(
+    pappl_printer_t *printer,		// I - Printer
+    pappl_job_t     *job,		// I - Current job or `NULL` if none
+    pappl_device_t  *device)		// I - Connection to device
+{
+  char			line[1025],	// Line from printer
+			*lineptr;	// Pointer into line
+  ssize_t		bytes;		// Bytes read
+  unsigned		errors = 0,	// Detected errors
+			warnings = 0;	// Detected warnings
+  pappl_preason_t	reasons;	// "printer-state-reasons" values
+
+
+  // Get the printer status...
+  if (papplDevicePuts(device, "~HQES\n") < 0)
+  {
+    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Unable to send HQES status command.");
+    return (false);
+  }
+
+  if ((bytes = papplDeviceRead(device, line, sizeof(line) - 1)) < 0)
+  {
+    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Unable to read HQES status response.");
+    return (false);
+  }
+
+  line[bytes] = '\0';
+
   papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "HQES returned '%s'.", line);
 
   if ((lineptr = strstr(line, "ERRORS:")) != NULL)
-    sscanf(lineptr + 7, "%*d%*x%x", &errors);
+  {
+    if (sscanf(lineptr + 7, "%*d%*x%x", &errors) != 1)
+      errors = 0;
+  }
+
   if ((lineptr = strstr(line, "WARNINGS:")) != NULL)
-    sscanf(lineptr + 9, "%*d%*x%x", &warnings);
+  {
+    if (sscanf(lineptr + 9, "%*d%*x%x", &warnings) != 1)
+      warnings = 0;
+  }
 
   reasons = PAPPL_PREASON_NONE;
 
@@ -854,70 +952,10 @@ lprint_zpl_status(
       papplLogPrinter(printer, PAPPL_LOGLEVEL_WARN, "At bin.");
   }
 
+  if (job && (reasons & PAPPL_PREASON_MEDIA_EMPTY))
+    reasons |= PAPPL_PREASON_MEDIA_NEEDED;
+
   papplPrinterSetReasons(printer, reasons, ~reasons);
 
-  // Query host status...
-  if (papplDevicePuts(device, "~HS\n") < 0)
-  {
-    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Unable to send HS status command.");
-    goto done;
-  }
-
-  if ((bytes = papplDeviceRead(device, line, sizeof(line) - 1)) < 0)
-  {
-    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Unable to read HS status response.");
-    goto done;
-  }
-
-  line[bytes] = '\0';
-
-  papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "HS returned '%s'.", line);
-
-  sscanf(line + 1, "%*d,%*d,%*d,%d", &length);
-  if (length > 11)
-  {
-    // Auto-detect label length for ready media...
-    pappl_pr_driver_data_t	data;	// Driver data
-
-    papplPrinterGetDriverData(printer, &data);
-
-    // 11 dots for the space between labels
-    length = 2540 * (length - 11) / data.y_resolution[0];
-
-    if (abs(length - 15240) <= 100)
-    {
-      // 4x6 label
-      papplCopyString(data.media_default.size_name, "na_index-4x6_4x6in", sizeof(data.media_default.size_name));
-      data.media_default.size_width  = 10160;
-      data.media_default.size_length = 15240;
-    }
-    else if (abs(length - 20320) <= 100)
-    {
-      // 4x8 label
-      papplCopyString(data.media_default.size_name, "oe_4x8-label_4x8in", sizeof(data.media_default.size_name));
-      data.media_default.size_width  = 10160;
-      data.media_default.size_length = 20320;
-    }
-    else
-    {
-      // *xN label
-      data.media_default.size_length = length;
-      snprintf(data.media_default.size_name, sizeof(data.media_default.size_name), "oe_%gx%g-label_%gx%gin", data.media_default.size_width / 2540.0, data.media_default.size_length / 2540.0, data.media_default.size_width / 2540.0, data.media_default.size_length / 2540.0);
-    }
-
-    if (data.media_default.size_width != data.media_ready[0].size_width || data.media_default.size_length != data.media_ready[0].size_length)
-    {
-      papplLogPrinter(printer, PAPPL_LOGLEVEL_INFO, "Label size changed to '%s'.", data.media_default.size_name);
-      papplPrinterSetDriverDefaults(printer, &data, 0, NULL);
-      papplPrinterSetReadyMedia(printer, 1, &data.media_default);
-    }
-  }
-
-  ret = true;
-
-  done:
-
-  papplPrinterCloseDevice(printer);
-
-  return (ret);
+  return (true);
 }
