@@ -1,37 +1,49 @@
 //
 // Main entry for LPrint, a Label Printer Application
 //
-// Copyright © 2019-2021 by Michael R Sweet.
+// Copyright © 2019-2024 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
 //
 
-//
-// Include necessary headers...
-//
-
 #include "lprint.h"
 #include "static-resources/lprint-de-strings.h"
 #include "static-resources/lprint-en-strings.h"
 #include "static-resources/lprint-es-strings.h"
 #include "static-resources/lprint-fr-strings.h"
 #include "static-resources/lprint-it-strings.h"
+#include "static-resources/lprint-css.h"
 #include "static-resources/lprint-png.h"
 #include "static-resources/lprint-large-png.h"
 #include "static-resources/lprint-small-png.h"
 
 
+//
+// Local types...
+//
+
+typedef struct lprint_device_s
+{
+  char	*device_info;			// Device description
+  char	*device_uri;			// Device URI
+  char	*device_id;			// Device ID
+} lprint_device_t;
+
+
 //
 // Local functions...
 //
 
-static const char *autoadd_cb(const char *device_info, const char *device_uri, const char *device_id, void *cbdata);
-static bool	driver_cb(pappl_system_t *system, const char *driver_name, const char *device_uri, const char *device_id, pappl_pr_driver_data_t *data, ipp_t **attrs, void *cbdata);
-static int	match_id(int num_did, cups_option_t *did, const char *match_id);
-static const char *mime_cb(const unsigned char *header, size_t headersize, void *data);
-static bool	printer_cb(const char *device_info, const char *device_uri, const char *device_id, pappl_system_t *system);
-static pappl_system_t *system_cb(int num_options, cups_option_t *options, void *data);
+static const char	*autoadd_cb(const char *device_info, const char *device_uri, const char *device_id, void *cbdata);
+static lprint_device_t	*copy_cb(lprint_device_t *src);
+static void		create_cb(pappl_printer_t *printer, void *cbdata);
+static bool		driver_cb(pappl_system_t *system, const char *driver_name, const char *device_uri, const char *device_id, pappl_pr_driver_data_t *data, ipp_t **attrs, void *cbdata);
+static void		free_cb(lprint_device_t *src);
+static int		match_id(int num_did, cups_option_t *did, const char *match_id);
+static const char	*mime_cb(const unsigned char *header, size_t headersize, void *data);
+static bool		printer_cb(const char *device_info, const char *device_uri, const char *device_id, cups_array_t *devices);
+static pappl_system_t	*system_cb(int num_options, cups_option_t *options, void *data);
 
 
 //
@@ -40,11 +52,19 @@ static pappl_system_t *system_cb(int num_options, cups_option_t *options, void *
 
 static pappl_pr_driver_t	lprint_drivers[] =
 {					// Driver list
+#ifdef LPRINT_EXPERIMENTAL
+#  include "lprint-brother.h"
+#  include "lprint-cpcl.h"
+#endif // LPRINT_EXPERIMENTAL
 #include "lprint-dymo.h"
 #include "lprint-epl2.h"
+#include "lprint-sii.h"
+#include "lprint-tspl.h"
 #include "lprint-zpl.h"
 };
-static char			lprint_statefile[1024];
+static char			lprint_spooldir[1024],
+					// Spool directory
+				lprint_statefile[1024];
 					// State file
 
 
@@ -58,7 +78,7 @@ main(int  argc,				// I - Number of command-line arguments
 {
   return (papplMainloop(argc, argv,
                         LPRINT_VERSION,
-                        "Copyright &copy; 2019-2021 by Michael R Sweet. All Rights Reserved.",
+                        "Copyright &copy; 2019-2024 by Michael R Sweet. All Rights Reserved.",
                         (int)(sizeof(lprint_drivers) / sizeof(lprint_drivers[0])),
                         lprint_drivers, autoadd_cb, driver_cb,
                         /*subcmd_name*/NULL, /*subcmd_cb*/NULL,
@@ -129,6 +149,63 @@ autoadd_cb(const char *device_info,	// I - Device information/name (not used)
 }
 
 
+//
+// 'copy_cb()' - Copy a device entry.
+//
+
+static lprint_device_t *		// O - New device entry
+copy_cb(lprint_device_t *src)		// I - Original device entry
+{
+  lprint_device_t	*dst;		// New device entry
+
+
+  if ((dst = (lprint_device_t *)calloc(1, sizeof(lprint_device_t))) != NULL)
+  {
+    dst->device_info = strdup(src->device_info);
+    dst->device_uri  = strdup(src->device_uri);
+    dst->device_id   = strdup(src->device_id);
+
+    if (!dst->device_info || !dst->device_uri || !dst->device_id)
+    {
+      free_cb(dst);
+      dst = NULL;
+    }
+  }
+
+  return (dst);
+}
+
+
+//
+// 'create_cb()' - Printer creation callback.
+//
+
+static void
+create_cb(pappl_printer_t *printer,	// I - Printer
+          void            *cbdata)	// I - Callback data (not used)
+{
+  char			resource[1024];	// Resource path
+  pappl_pr_driver_data_t data;		// Driver data
+
+
+  (void)cbdata;
+
+  LPRINT_DEBUG("create_cb(printer=%p(%s), cbdata=%p)\n", printer, printer ? papplPrinterGetName(printer) : "null", cbdata);
+
+  // Add custom label media page to replace the standard ready media page...
+  papplSystemRemoveResource(papplPrinterGetSystem(printer), papplPrinterGetPath(printer, "media", resource, sizeof(resource)));
+  papplSystemAddResourceCallback(papplPrinterGetSystem(printer), resource, "text/html", (pappl_resource_cb_t)lprintMediaUI, printer);
+  LPRINT_DEBUG("create_cb: Added new media page for '%s'.\n", resource);
+
+  // Load custom media sizes and report them...
+  papplPrinterGetDriverData(printer, &data);
+  lprintMediaLoad(printer, &data);
+  lprintMediaUpdate(printer, &data);
+  papplPrinterSetDriverData(printer, &data, NULL);
+
+  LPRINT_DEBUG("create_cb: data.extension=%p\n", data.extension);
+}
+
 
 //
 // 'driver_cb()' - Main driver callback.
@@ -144,6 +221,7 @@ driver_cb(
     ipp_t                  **attrs,	// O - Pointer to driver attributes
     void                   *cbdata)	// I - Callback data (not used)
 {
+  bool	ret = false;			// Return value
   int	i;				// Looping var
 
 
@@ -168,10 +246,9 @@ driver_cb(
   data->kind = PAPPL_KIND_LABEL;
 
   // Color values...
-  data->color_supported   = PAPPL_COLOR_MODE_AUTO | PAPPL_COLOR_MODE_MONOCHROME;
-  data->color_default     = PAPPL_COLOR_MODE_MONOCHROME;
-  data->raster_types      = PAPPL_PWG_RASTER_TYPE_BLACK_1 | PAPPL_PWG_RASTER_TYPE_BLACK_8 | PAPPL_PWG_RASTER_TYPE_SGRAY_8;
-  data->force_raster_type = PAPPL_PWG_RASTER_TYPE_BLACK_1;
+  data->color_supported = PAPPL_COLOR_MODE_AUTO | PAPPL_COLOR_MODE_MONOCHROME | PAPPL_COLOR_MODE_BI_LEVEL;
+  data->color_default   = PAPPL_COLOR_MODE_MONOCHROME;
+  data->raster_types    = PAPPL_PWG_RASTER_TYPE_BLACK_1 | PAPPL_PWG_RASTER_TYPE_BLACK_8 | PAPPL_PWG_RASTER_TYPE_SGRAY_8;
 
   // "print-quality-default" value...
   data->quality_default = IPP_QUALITY_NORMAL;
@@ -199,14 +276,57 @@ driver_cb(
   data->testpage_cb = lprintTestPageCB;
 
   // Use the corresponding sub-driver callback to set things up...
-  if (!strncmp(driver_name, "dymo_", 5))
-    return (lprintDYMO(system, driver_name, device_uri, device_id, data, attrs, cbdata));
-  else if (!strncmp(driver_name, "epl2_", 5))
-    return (lprintEPL2(system, driver_name, device_uri, device_id, data, attrs, cbdata));
-  else if (!strncmp(driver_name, "zpl_", 4))
-    return (lprintZPL(system, driver_name, device_uri, device_id, data, attrs, cbdata));
+#ifdef LPRINT_EXPERIMENTAL
+  if (!strncmp(driver_name, "brother_", 8))
+    ret = lprintBrother(system, driver_name, device_uri, device_id, data, attrs, cbdata);
   else
-    return (false);
+#endif // LPRINT_EXPERIMENTAL
+  if (!strncmp(driver_name, "dymo_", 5))
+    ret = lprintDYMO(system, driver_name, device_uri, device_id, data, attrs, cbdata);
+  else if (!strncmp(driver_name, "epl2_", 5))
+    ret = lprintEPL2(system, driver_name, device_uri, device_id, data, attrs, cbdata);
+  else if (!strncmp(driver_name, "sii_", 4))
+    ret = lprintSII(system, driver_name, device_uri, device_id, data, attrs, cbdata);
+  else if (!strncmp(driver_name, "tspl_", 5))
+    ret = lprintTSPL(system, driver_name, device_uri, device_id, data, attrs, cbdata);
+  else if (!strncmp(driver_name, "zpl_", 4))
+    ret = lprintZPL(system, driver_name, device_uri, device_id, data, attrs, cbdata);
+
+  // Update the ready media...
+  for (i = 0; i < data->num_source; i ++)
+  {
+    pwg_media_t *pwg = pwgMediaForPWG(data->media_ready[i].size_name);
+
+    data->media_ready[i].bottom_margin = data->bottom_top;
+    data->media_ready[i].left_margin   = data->left_right;
+    data->media_ready[i].right_margin  = data->left_right;
+    data->media_ready[i].size_width    = pwg->width;
+    data->media_ready[i].size_length   = pwg->length;
+    data->media_ready[i].top_margin    = data->bottom_top;
+    papplCopyString(data->media_ready[i].source, data->source[i], sizeof(data->media_ready[i].source));
+    if (!data->media_ready[i].type[0])
+      papplCopyString(data->media_ready[i].type, data->type[0], sizeof(data->media_ready[i].type));
+  }
+
+  // By default use media from the main source...
+  data->media_default = data->media_ready[0];
+
+  // Return...
+  return (ret);
+}
+
+
+//
+// 'free_cb()' - Free a device entry.
+//
+
+static void
+free_cb(lprint_device_t *dev)		// I - Device entry
+{
+  free(dev->device_info);
+  free(dev->device_uri);
+  free(dev->device_id);
+  free(dev);
 }
 
 
@@ -296,6 +416,10 @@ mime_cb(const unsigned char *header,	// I - Header data
 
   if (headersize >= sizeof(testpage) && !memcmp(header, testpage, sizeof(testpage)))
     return (LPRINT_TESTPAGE_MIMETYPE);
+  else if (headersize >= 2 && header[0] == '^' && isupper(header[1] & 255))
+    return (LPRINT_ZPL_MIMETYPE);
+  else if (headersize >= 3 && !memcmp(header, "\nN\n", 3))
+    return (LPRINT_EPL2_MIMETYPE);
   else
     return (NULL);
 }
@@ -306,46 +430,19 @@ mime_cb(const unsigned char *header,	// I - Header data
 //
 
 static bool				// O - `false` to continue
-printer_cb(const char     *device_info,	// I - Device information
-	   const char     *device_uri,	// I - Device URI
-	   const char     *device_id,	// I - IEEE-1284 device ID
-	   pappl_system_t *system)	// I - System
+printer_cb(const char   *device_info,	// I - Device information
+	   const char   *device_uri,	// I - Device URI
+	   const char   *device_id,	// I - IEEE-1284 device ID
+	   cups_array_t *devices)	// I - Device array
 {
-  const char *driver_name = autoadd_cb(device_info, device_uri, device_id, system);
-					// Driver name, if any
-
-  if (driver_name)
-  {
-    char	name[1024],		// Printer name
-		*nameptr;		// Pointer in name
+  lprint_device_t	dev;		// New device
 
 
-    // Zebra puts "Zebra Technologies ZTC" on the front of their printer names,
-    // which is a bit, um, wordy.  Clean up the device info string to use as a
-    // printer name and drop any trailing "(ID)" nonsense if we don't need it.
-    if (!strncasecmp(device_info, "Zebra Technologies ZTC ", 23))
-      snprintf(name, sizeof(name), "Zebra %s", device_info + 23);
-    else
-      papplCopyString(name, device_info, sizeof(name));
+  dev.device_info = (char *)device_info;
+  dev.device_uri  = (char *)device_uri;
+  dev.device_id   = (char *)device_id;
 
-    if ((nameptr = strstr(name, " (")) != NULL)
-      *nameptr = '\0';
-
-    if (!papplPrinterCreate(system, 0, name, driver_name, device_id, device_uri))
-    {
-      // Printer already exists with this name, so try adding a number to the
-      // name...
-      int	i;			// Looping var
-      char	newname[1024];		// New name
-
-      for (i = 2; i < 100; i ++)
-      {
-        snprintf(newname, sizeof(newname), "%s %d", name, i);
-        if (papplPrinterCreate(system, 0, name, driver_name, device_id, device_uri))
-          break;
-      }
-    }
-  }
+  cupsArrayAdd(devices, &dev);
 
   return (false);
 }
@@ -364,7 +461,9 @@ system_cb(
   pappl_system_t	*system;	// System object
   const char		*val,		// Current option value
 			*hostname,	// Hostname, if any
+			*listenhost,	// Listen hostname, if any
 			*logfile,	// Log file, if any
+			*spooldir,	// Spool directory, if any
 			*system_name;	// System name, if any
   pappl_loglevel_t	loglevel;	// Log level
   int			port = 0;	// Port number, if any
@@ -378,10 +477,6 @@ system_cb(
 
   (void)data;
 
-  // Enable/disable TLS...
-  if ((val = cupsGetOption("tls", num_options, options)) != NULL && !strcmp(val, "no"))
-    soptions |= PAPPL_SOPTIONS_NO_TLS;
-
   // Parse standard log and server options...
   if ((val = cupsGetOption("log-level", num_options, options)) != NULL)
   {
@@ -404,8 +499,44 @@ system_cb(
   else
     loglevel = PAPPL_LOGLEVEL_UNSPEC;
 
-  logfile     = cupsGetOption("log-file", num_options, options);
+  if ((val = cupsGetOption("server-options", num_options, options)) != NULL)
+  {
+    const char	*valptr;		// Pointer into value
+
+    for (valptr = val; valptr && *valptr;)
+    {
+      if (!strcmp(valptr, "none") || !strncmp(valptr, "none,", 5))
+        soptions = PAPPL_SOPTIONS_NONE;
+      else if (!strcmp(valptr, "dnssd-host") || !strncmp(valptr, "dnssd-host,", 11))
+        soptions |= PAPPL_SOPTIONS_DNSSD_HOST;
+      else if (!strcmp(valptr, "no-multi-queue") || !strncmp(valptr, "no-multi-queue,", 15))
+        soptions &= (pappl_soptions_t)~PAPPL_SOPTIONS_MULTI_QUEUE;
+      else if (!strcmp(valptr, "raw-socket") || !strncmp(valptr, "raw-socket,", 11))
+        soptions |= PAPPL_SOPTIONS_RAW_SOCKET;
+      else if (!strcmp(valptr, "usb-printer") || !strncmp(valptr, "usb-printer,", 12))
+        soptions |= PAPPL_SOPTIONS_USB_PRINTER;
+      else if (!strcmp(valptr, "no-web-interface") || !strncmp(valptr, "no-web-interface,", 17))
+        soptions &= (pappl_soptions_t)~PAPPL_SOPTIONS_WEB_INTERFACE;
+      else if (!strcmp(valptr, "web-log") || !strncmp(valptr, "web-log,", 8))
+        soptions |= PAPPL_SOPTIONS_WEB_LOG;
+      else if (!strcmp(valptr, "web-network") || !strncmp(valptr, "web-network,", 12))
+        soptions |= PAPPL_SOPTIONS_WEB_NETWORK;
+      else if (!strcmp(valptr, "web-remote") || !strncmp(valptr, "web-remote,", 11))
+        soptions |= PAPPL_SOPTIONS_WEB_REMOTE;
+      else if (!strcmp(valptr, "web-security") || !strncmp(valptr, "web-security,", 13))
+        soptions |= PAPPL_SOPTIONS_WEB_SECURITY;
+      else if (!strcmp(valptr, "no-tls") || !strncmp(valptr, "no-tls,", 7))
+        soptions |= PAPPL_SOPTIONS_NO_TLS;
+
+      if ((valptr = strchr(valptr, ',')) != NULL)
+        valptr ++;
+    }
+  }
+
   hostname    = cupsGetOption("server-hostname", num_options, options);
+  listenhost  = cupsGetOption("listen-hostname", num_options, options);
+  logfile     = cupsGetOption("log-file", num_options, options);
+  spooldir    = cupsGetOption("spool-directory", num_options, options);
   system_name = cupsGetOption("system-name", num_options, options);
 
   if ((val = cupsGetOption("server-port", num_options, options)) != NULL)
@@ -419,10 +550,14 @@ system_cb(
       port = atoi(val);
   }
 
-  // State file...
+  // Spool directory and state file...
   if ((val = getenv("SNAP_DATA")) != NULL)
   {
+    snprintf(lprint_spooldir, sizeof(lprint_spooldir), "%s/lprint.d", val);
     snprintf(lprint_statefile, sizeof(lprint_statefile), "%s/lprint.conf", val);
+
+    if (!spooldir)
+      spooldir = lprint_spooldir;
   }
   else if ((val = getenv("XDG_DATA_HOME")) != NULL)
   {
@@ -431,7 +566,11 @@ system_cb(
 #ifdef _WIN32
   else if ((val = getenv("USERPROFILE")) != NULL)
   {
-    snprintf(lprint_statefile, sizeof(lprint_statefile), "%s/AppData/Local/lprint.conf", val);
+    snprintf(lprint_spooldir, sizeof(lprint_spooldir), "%s/AppData/Local/lprint.d", val);
+    snprintf(lprint_statefile, sizeof(lprint_statefile), "%s/AppData/Local/lprint.ini", val);
+
+    if (!spooldir)
+      spooldir = lprint_spooldir;
   }
   else
   {
@@ -440,7 +579,15 @@ system_cb(
 #else
   else if ((val = getenv("HOME")) != NULL)
   {
+#  ifdef __APPLE__
+    snprintf(lprint_spooldir, sizeof(lprint_spooldir), "%s/Library/Application Support/lprint.d", val);
+    snprintf(lprint_statefile, sizeof(lprint_statefile), "%s/Library/Application Support/lprint.state", val);
+
+    if (!spooldir)
+      spooldir = lprint_spooldir;
+#  else
     snprintf(lprint_statefile, sizeof(lprint_statefile), "%s/.lprint.conf", val);
+#  endif // __APPLE__
   }
   else
   {
@@ -448,39 +595,115 @@ system_cb(
   }
 #endif // _WIN32
 
+  if (spooldir && access(spooldir, 0))
+  {
+    if (mkdir(spooldir, 0777))
+    {
+      perror(spooldir);
+      return (NULL);
+    }
+  }
+
   // Create the system object...
-  if ((system = papplSystemCreate(soptions, system_name ? system_name : "LPrint", port, "_print,_universal", cupsGetOption("spool-directory", num_options, options), logfile ? logfile : "-", loglevel, cupsGetOption("auth-service", num_options, options), /* tls_only */false)) == NULL)
+  if ((system = papplSystemCreate(soptions, system_name ? system_name : "LPrint", port, "_print,_universal", spooldir, logfile ? logfile : "-", loglevel, cupsGetOption("auth-service", num_options, options), /* tls_only */false)) == NULL)
     return (NULL);
 
-  papplSystemAddListeners(system, NULL);
+  if (!cupsGetOption("private-server", num_options, options))
+  {
+    // Listen for TCP/IP connections...
+    papplSystemAddListeners(system, listenhost);
+  }
+
   papplSystemSetHostName(system, hostname);
 
+  if ((val = cupsGetOption("admin-group", num_options, options)) != NULL)
+    papplSystemSetAdminGroup(system, val);
+
   papplSystemSetMIMECallback(system, mime_cb, NULL);
   papplSystemAddMIMEFilter(system, LPRINT_TESTPAGE_MIMETYPE, "image/pwg-raster", lprintTestFilterCB, NULL);
 
-  papplSystemSetPrinterDrivers(system, (int)(sizeof(lprint_drivers) / sizeof(lprint_drivers[0])), lprint_drivers, autoadd_cb, /*create_cb*/NULL, driver_cb, system);
+  papplSystemSetPrinterDrivers(system, (int)(sizeof(lprint_drivers) / sizeof(lprint_drivers[0])), lprint_drivers, autoadd_cb, create_cb, driver_cb, system);
 
   papplSystemAddResourceData(system, "/favicon.png", "image/png", lprint_small_png, sizeof(lprint_small_png));
   papplSystemAddResourceData(system, "/navicon.png", "image/png", lprint_png, sizeof(lprint_png));
-  papplSystemAddResourceString(system, "/de.strings", "text/strings", lprint_de_strings);
-  papplSystemAddResourceString(system, "/en.strings", "text/strings", lprint_en_strings);
-  papplSystemAddResourceString(system, "/es.strings", "text/strings", lprint_es_strings);
-  papplSystemAddResourceString(system, "/fr.strings", "text/strings", lprint_fr_strings);
-  papplSystemAddResourceString(system, "/it.strings", "text/strings", lprint_it_strings);
+  papplSystemAddResourceString(system, "/style.css", "text/css", lprint_css);
+  papplSystemAddStringsData(system, "/de.strings", "de", lprint_de_strings);
+  papplSystemAddStringsData(system, "/en.strings", "en", lprint_en_strings);
+  papplSystemAddStringsData(system, "/es.strings", "es", lprint_es_strings);
+  papplSystemAddStringsData(system, "/fr.strings", "fr", lprint_fr_strings);
+  papplSystemAddStringsData(system, "/it.strings", "it", lprint_it_strings);
 
-  papplSystemSetFooterHTML(system, "Copyright &copy; 2019-2021 by Michael R Sweet. All rights reserved.");
+  papplSystemSetFooterHTML(system, "Copyright &copy; 2019-2024 by Michael R Sweet. All rights reserved.");
   papplSystemSetSaveCallback(system, (pappl_save_cb_t)papplSystemSaveState, (void *)lprint_statefile);
   papplSystemSetVersions(system, (int)(sizeof(versions) / sizeof(versions[0])), versions);
 
-  fprintf(stderr, "lprint: statefile='%s'\n", lprint_statefile);
-
   if (!papplSystemLoadState(system, lprint_statefile))
   {
     // No old state, use defaults and auto-add printers...
+    cups_array_t	*devices;	// Device array
+    lprint_device_t	*dev;		// Current device
+
     papplSystemSetDNSSDName(system, system_name ? system_name : "LPrint");
 
     papplLog(system, PAPPL_LOGLEVEL_INFO, "Auto-adding printers...");
-    papplDeviceList(PAPPL_DEVTYPE_USB, (pappl_device_cb_t)printer_cb, system, papplLogDevice, system);
+    devices = cupsArrayNew(NULL, NULL, NULL, 0, (cups_acopy_cb_t)copy_cb, (cups_afree_cb_t)free_cb);
+    papplDeviceList(PAPPL_DEVTYPE_USB, (pappl_device_cb_t)printer_cb, devices, papplLogDevice, system);
+
+    for (dev = (lprint_device_t *)cupsArrayGetFirst(devices); dev; dev = (lprint_device_t *)cupsArrayGetNext(devices))
+    {
+      const char *driver_name = autoadd_cb(dev->device_info, dev->device_uri, dev->device_id, system);
+					// Driver name, if any
+
+      printf("%s (%s;%s) -> %s\n", dev->device_info, dev->device_uri, dev->device_id, driver_name);
+
+      if (driver_name)
+      {
+	char	name[128],		// Printer name
+		*nameptr;		// Pointer in name
+
+	// Zebra puts "Zebra Technologies ZTC" on the front of their printer names,
+	// which is a bit, um, wordy.  Clean up the device info string to use as a
+	// printer name and drop any trailing "(ID)" nonsense if we don't need it.
+	if (!strncasecmp(dev->device_info, "Zebra Technologies ZTC ", 23))
+	  snprintf(name, sizeof(name), "Zebra %s", dev->device_info + 23);
+	else
+	  papplCopyString(name, dev->device_info, sizeof(name));
+
+	if ((nameptr = strstr(name, " (")) != NULL)
+	  *nameptr = '\0';
+
+	if (!papplPrinterCreate(system, 0, name, driver_name, dev->device_id, dev->device_uri))
+	{
+	  // Printer already exists with this name, so try adding a number to the
+	  // name...
+	  int	i;			// Looping var
+	  char	newname[128],		// New name
+		number[4];		// Number string
+	  size_t namelen = strlen(name),// Length of original name string
+		numberlen;		// Length of number string
+
+	  for (i = 2; i < 100; i ++)
+	  {
+	    // Append " NNN" to the name, truncating the existing name as needed to
+	    // include the number at the end...
+	    snprintf(number, sizeof(number), " %d", i);
+	    numberlen = strlen(number);
+
+	    papplCopyString(newname, name, sizeof(newname));
+	    if ((namelen + numberlen) < sizeof(newname))
+	      memcpy(newname + namelen, number, numberlen + 1);
+	    else
+	      memcpy(newname + sizeof(newname) - numberlen - 1, number, numberlen + 1);
+
+	    // Try creating with this name...
+	    if (papplPrinterCreate(system, 0, newname, driver_name, dev->device_id, dev->device_uri))
+	      break;
+	  }
+	}
+      }
+    }
+
+    cupsArrayDelete(devices);
   }
 
   return (system);
