@@ -1,14 +1,19 @@
 //
 // Xcode configuration header file for LPrint, a Label Printer Application
 //
-// Copyright © 2019-2021 by Michael R Sweet.
+// Copyright © 2019-2024 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
 //
 
 // Version number
-#define LPRINT_VERSION "1.1.0"
+#define LPRINT_VERSION "1.3.1"
 #define LPRINT_MAJOR_VERSION 1
-#define LPRINT_MINOR_VERSION 1
-#define LPRINT_PATCH_VERSION 0
+#define LPRINT_MINOR_VERSION 3
+#define LPRINT_PATCH_VERSION 1
+
+
+// Enable experimental drivers?
+/*#undef LPRINT_EXPERIMENTAL */
+#define LPRINT_EXPERIMENTAL 1
